from .base import Entity
from typing import *

class ForwardDeclaration: ...

class RelationshipDeclaration:
    def __init__(self, parent: Union[Type['Entity'], Type[ForwardDeclareCompare]], child: Union[Type['Entity'], Type[ForwardDeclareCompare], str]): ...

class ForwardDeclareCompare(type):
    def __eq__(self, other: Union[Type['Entity'], Type[ForwardDeclaration]]) -> bool: ...

class Asset(ForwardDeclaration): ...
class Scope(ForwardDeclaration): ...
class Metadata(ForwardDeclaration): ...
class TaskTemplateItem(ForwardDeclaration): ...
class ReviewSessionInvitee(ForwardDeclaration): ...
class Bid(ForwardDeclaration): ...
class Milestone(ForwardDeclaration): ...
class SequenceComponent(ForwardDeclaration): ...
class ObjectType(ForwardDeclaration): ...
class ReviewSessionObjectStatus(ForwardDeclaration): ...
class Task(ForwardDeclaration): ...
class Project(ForwardDeclaration): ...
class Conversation(ForwardDeclaration): ...
class AssetVersionLink(ForwardDeclaration): ...
class Timelog(ForwardDeclaration): ...
class Shot(ForwardDeclaration): ...
class ListCustomAttributeValue(ForwardDeclaration): ...
class Disk(ForwardDeclaration): ...
class Event(ForwardDeclaration): ...
class Participant(ForwardDeclaration): ...
class JobComponent(ForwardDeclaration): ...
class NoteCategory(ForwardDeclaration): ...
class State(ForwardDeclaration): ...
class Setting(ForwardDeclaration): ...
class Location(ForwardDeclaration): ...
class Episode(ForwardDeclaration): ...
class CalendarEvent(ForwardDeclaration): ...
class Type(ForwardDeclaration): ...
class ContextCustomAttributeValue(ForwardDeclaration): ...
class Job(ForwardDeclaration): ...
class TypedContextLink(ForwardDeclaration): ...
class Chapter(ForwardDeclaration): ...
class AssetType(ForwardDeclaration): ...
class Timer(ForwardDeclaration): ...
class CustomAttributeGroup(ForwardDeclaration): ...
class ReviewSession(ForwardDeclaration): ...
class AssetVersion(ForwardDeclaration): ...
class AssetBuild(ForwardDeclaration): ...
class UserCustomAttributeValue(ForwardDeclaration): ...
class ContainerComponent(ForwardDeclaration): ...
class Message(ForwardDeclaration): ...
class CustomAttributeValue(ForwardDeclaration): ...
class TypedContext(ForwardDeclaration): ...
class Group(ForwardDeclaration): ...
class Note(ForwardDeclaration): ...
class WorkflowSchema(ForwardDeclaration): ...
class DashboardWidget(ForwardDeclaration): ...
class Epic(ForwardDeclaration): ...
class Status(ForwardDeclaration): ...
class Appointment(ForwardDeclaration): ...
class CalendarEventResource(ForwardDeclaration): ...
class Membership(ForwardDeclaration): ...
class User(ForwardDeclaration): ...
class ProjectSchemaOverride(ForwardDeclaration): ...
class Recipient(ForwardDeclaration): ...
class TypedContextStatusChange(ForwardDeclaration): ...
class AssetVersionCustomAttributeValue(ForwardDeclaration): ...
class ProjectSchema(ForwardDeclaration): ...
class TaskTemplate(ForwardDeclaration): ...
class EntitySetting(ForwardDeclaration): ...
class ListCategory(ForwardDeclaration): ...
class StatusChange(ForwardDeclaration): ...
class Feed(ForwardDeclaration): ...
class Priority(ForwardDeclaration): ...
class AssetCustomAttributeValue(ForwardDeclaration): ...
class NoteComponent(ForwardDeclaration): ...
class TaskTypeSchema(ForwardDeclaration): ...
class TypedContextList(ForwardDeclaration): ...
class AssetVersionStatusChange(ForwardDeclaration): ...
class UserSecurityRole(ForwardDeclaration): ...
class FileComponent(ForwardDeclaration): ...
class AssetGroup(ForwardDeclaration): ...
class Folder(ForwardDeclaration): ...
class Schema(ForwardDeclaration): ...
class CustomAttributeType(ForwardDeclaration): ...
class Component(ForwardDeclaration): ...
class SchemaType(ForwardDeclaration): ...
class SchemaStatus(ForwardDeclaration): ...
class Resource(ForwardDeclaration): ...
class SecurityRole(ForwardDeclaration): ...
class List(ForwardDeclaration): ...
class ReviewSessionObject(ForwardDeclaration): ...
class CustomAttributeConfiguration(ForwardDeclaration): ...
class Context(ForwardDeclaration): ...
class AssetVersionList(ForwardDeclaration): ...
class Sequence(ForwardDeclaration): ...
class ComponentLocation(ForwardDeclaration): ...
class Dashboard(ForwardDeclaration): ...
